@@ -24,6 +24,9 @@ import tempfile
 import io
 import gi
 import locale
+import packaging.version as version
+from typing import Any, Dict, List
+
 from . import metadata
 from gi.repository import Gtk
 gi.require_version("Poppler", "0.18")
@@ -31,6 +34,8 @@ from gi.repository import Poppler
 import gettext
 _ = gettext.gettext
 
+from .core import Page, Sides
+
 # pikepdf.Page.add_overlay()/add_underlay() can't place a page exactly
 # if for example LC_NUMERIC=fi_FI
 try:
@@ -39,30 +44,31 @@ except locale.Error:
     pass  # Gtk already prints a warning
 
 
-def layer_support():
-    """Pikepdf >= 3 has overlay/underlay support in pdfarranger."""
-    layer_support = True
-    pdf1 = pikepdf.Pdf.new()
-    pdf1.add_blank_page()
-    pdf2 = pikepdf.Pdf.new()
-    pdf2.add_blank_page()
-    try:
-        fpage = pdf1.copy_foreign(pdf2.pages[0])
-    except NotImplementedError:
-        # This is pikepdf >= 6
-        pass
-    else:
-        try:
-            pdf1.pages[0].add_overlay(fpage)
-        except AttributeError:
-            # This is pikepdf < 3
-            layer_support = False
-    pdf1.close()
-    pdf2.close()
-    return layer_support
+def get_blank_doc(pageadder, pdfqueue, tmpdir, size, npages=1):
+    """Search pdfqueue for a matching pdf with blank pages. Create it if it does not exist.
+
+    Some notes:
+    Blank pdf documents are created prior to export (vs created as needed at export)
+    because it will keep code simpler. For example rendering of thumbnails require no
+    extra for rendering of a blank page.
+
+    A document with several blank pages is needed if the page number under thumbnail
+    need to be something else than 1.
+    """
+    for i, pdfdoc in enumerate(pdfqueue):
+        if size == pdfdoc.blank_size and npages == pdfdoc.document.get_n_pages():
+            filename = pdfdoc.copyname
+            nfile = i + 1
+            return filename, nfile
+    filename = _create_blank_page(tmpdir, size, npages)
+    doc_data = pageadder.get_pdfdoc(filename, basename=None, blank_size=size)
+    if doc_data is None:
+        return None, None
+    nfile = doc_data[1]
+    return filename, nfile
 
 
-def create_blank_page(tmpdir, size, npages=1):
+def _create_blank_page(tmpdir, size, npages=1):
     """
     Create a temporary PDF file with npages empty pages.
     The size is in PDF unit (1/72 of inch).
@@ -107,7 +113,7 @@ def _intersect_rectangle(rect1, rect2):
     ]
 
 
-def _mediabox(page, crop=None):
+def _mediabox(page, crop=Sides()):
     """ Return the media box for a given page. """
     # PDF files which do not have mediabox default to Portrait Letter / ANSI A
     cmb = page.MediaBox if "/MediaBox" in page else [0, 0, 612, 792]
@@ -117,7 +123,7 @@ def _mediabox(page, crop=None):
         # reduced to their intersection with the media box"
         cmb = _intersect_rectangle(cmb, _normalize_rectangle(page.CropBox))
 
-    if crop is None or crop == [0., 0., 0., 0.]:
+    if crop == Sides():
         return cmb
     angle = page.Rotate if '/Rotate' in page else 0
     rotate_times = int(round(((angle) % 360) / 90) % 4)
@@ -127,12 +133,12 @@ def _mediabox(page, crop=None):
         for _ in range(rotate_times):
             perm.append(perm.pop(0))
         perm.insert(1, perm.pop(2))
-        crop = [crop_init[perm[side]] for side in range(4)]
+        crop = Sides(*(crop_init[perm[side]] for side in range(4)))
     x1, y1, x2, y2 = [float(x) for x in cmb]
-    x1_new = x1 + (x2 - x1) * crop[0]
-    x2_new = x2 - (x2 - x1) * crop[1]
-    y1_new = y1 + (y2 - y1) * crop[3]
-    y2_new = y2 - (y2 - y1) * crop[2]
+    x1_new = x1 + (x2 - x1) * crop.left
+    x2_new = x2 - (x2 - x1) * crop.right
+    y1_new = y1 + (y2 - y1) * crop.bottom
+    y2_new = y2 - (y2 - y1) * crop.top
     return [x1_new, y1_new, x2_new, y2_new]
 
 
@@ -170,39 +176,10 @@ def _scale(doc, page, factor):
     # It's also needed with pikepdf 4.2 else we get:
     # RuntimeError: QPDFPageObjectHelper::getFormXObjectForPage called with a direct object
     # when calling as_form_xobject in generate_booklet
+    # It's also needed for pikepdf 8.8 but not for pikepdf 6.0.1
     new_page = doc.make_indirect(new_page)
     return new_page
 
-def check_content(parent, pdf_list):
-    """ Warn about fillable forms or outlines that are lost on export."""
-    warn = False
-    for pdf in [pikepdf.open(p.copyname, password=p.password) for p in pdf_list]:
-        if "/AcroForm" in pdf.Root.keys(): # fillable form
-            warn = True
-            break
-        if pdf.open_outline().root: # table of contents
-            warn = True
-            break
-    if warn:
-        d = Gtk.Dialog(_('Warning'),
-                       parent=parent,
-                       flags=Gtk.DialogFlags.MODAL,
-                       buttons=("_Cancel", Gtk.ResponseType.CANCEL,
-                                "_OK", Gtk.ResponseType.OK))
-        label = Gtk.Label(_('Forms and outlines are lost on saving.'))
-        d.vbox.pack_start(label, False, False, 6)
-        checkbutton = Gtk.CheckButton(_('Do not show this dialog again.'))
-        d.vbox.pack_start(checkbutton, False, False, 6)
-        buttonbox = d.get_action_area()
-        buttons = buttonbox.get_children()
-        d.set_focus(buttons[1])
-        d.show_all()
-        response = d.run()
-        enable_warnings = not checkbutton.get_active()
-        d.destroy()
-        return response, enable_warnings
-    return Gtk.ResponseType.OK, True
-
 
 def _update_angle(model_page, source_page, output_page):
     angle = model_page.angle
@@ -222,7 +199,22 @@ def _apply_geom_transform(pdf_output, new_page, row):
         del new_page.TrimBox
     if '/CropBox' in new_page:
         del new_page.CropBox
-    return _scale(pdf_output, new_page, row.scale)
+    return pikepdf.Page(_scale(pdf_output, new_page, row.scale))
+
+
+def _apply_geom_transform_job(pdf_output:pikepdf.Pdf, new_page:pikepdf.Page, page:Page) -> None:
+    new_page.rotate(page.angle, relative=True)
+    new_page.MediaBox = _mediabox(new_page, page.crop)
+    # add_overlay() & add_underlay() will use TrimBox or CropBox if they exist
+    if '/TrimBox' in new_page:
+        del new_page.TrimBox
+    if '/CropBox' in new_page:
+        del new_page.CropBox
+    if page.scale != 1:
+        pdf_output.pages.append(new_page)
+        new_page.obj.emplace(_scale(pdf_output, pdf_output.pages[-1], page.scale))
+        del(pdf_output.pages[-1])
+
 
 
 def _remove_unreferenced_resources(pdfdoc):
@@ -308,7 +300,7 @@ def _copy_n_transform(pdf_input, pdf_output, pages, quit_flag=None):
         rotate_times = int(round(((dangle0) % 360) / 90) % 4)
         for lprow in row.layerpages:
             # Rotate the offsets so they are relative to dest page
-            offset = lprow.rotate_array(lprow.offset, rotate_times)
+            offset = lprow.offset.rotated(rotate_times)
             offs_left, offs_right, offs_top, offs_bottom = offset
             x1 = row.scale * (dx1 + dw * offs_left)
             y1 = row.scale * (dy1 + dh * offs_bottom)
@@ -329,12 +321,7 @@ def _append_page(current_page, copied_pages, pdf_output, row):
     """Add a page to the output pdf. A page that already exist is duplicated."""
     new_page = copied_pages.get((row.nfile, row.npage))
     if new_page is None:
-        try:
-            # for backward compatibility with pikepdf <= 3
-            new_page = pdf_output.copy_foreign(current_page)
-        except NotImplementedError:
-            # This is pikepdf >= 6
-            new_page = current_page
+        new_page = current_page
     # let pdf_output adopt new_page
     pdf_output.pages.append(new_page)
     new_page = pdf_output.pages[-1]
@@ -348,7 +335,53 @@ def _append_page(current_page, copied_pages, pdf_output, row):
         new_page.Annots = pdf_output.copy_foreign(indirect_annots)
 
 
-def export_doc(pdf_input, pages, mdata, files_out, quit_flag):
+def _transform_job(pdf_output: pikepdf.Pdf, pages: List[Page], quit_flag = None) -> None:
+    """ Same as _copy_n_transform, except it doesn't copy. Requires pikepdf >= 8.0 """
+    # Fix missing MediaBoxes
+    for page in pdf_output.pages:
+        if page.mediabox is None:
+            page.mediabox = pikepdf.Array((0, 0, 612, 792))
+
+    # We don't need to call _append_page as the Job interface copies pages / annotations as necessary.
+    mediaboxes:List[pikepdf.Rectangle] = []
+    i = 0
+    for page in pages:
+        if quit_flag is not None and quit_flag.is_set():
+            return
+        mediaboxes.append(pikepdf.Rectangle(pdf_output.pages[i].mediabox))
+        _apply_geom_transform_job(pdf_output, pdf_output.pages[i], page)
+        for lpage in page.layerpages:
+            i += 1
+            _apply_geom_transform_job(pdf_output, pdf_output.pages[i], lpage)
+        i += 1
+
+    # # Add overlays and underlays
+    for i, page in enumerate(pages):
+        # The dest page coordinates and size before geometrical transformations
+        mb = mediaboxes[i]
+
+        # Call to rotate in _apply_geom_transform_job ensures /Rotate exists
+        rotate_times = int(round((pdf_output.pages[i].Rotate % 360) / 90) % 4)
+        for lpage in page.layerpages:
+            # Rotate the offsets so they are relative to dest page
+            offset = lpage.offset.rotated(rotate_times)
+            offs_left, offs_right, offs_top, offs_bottom = offset
+            x1 = page.scale * (mb.llx + mb.width * offs_left)
+            y1 = page.scale * (mb.lly + mb.height * offs_bottom)
+            x2 = page.scale * (mb.llx + mb.width * (1 - offs_right))
+            y2 = page.scale * (mb.lly + mb.height * (1 - offs_top))
+            rect = pikepdf.Rectangle(x1, y1, x2, y2)
+
+            if lpage.laypos == 'OVERLAY':
+                pdf_output.pages[i].add_overlay(pdf_output.pages[i + 1], rect)
+            else:
+                pdf_output.pages[i].add_underlay(pdf_output.pages[i + 1], rect)
+            # Remove the temporary added page
+            del pdf_output.pages[i + 1]
+
+
+
+def export_doc(pdf_input, pages, mdata, files_out, quit_flag, test_mode=False):
     """Same as export() but with pikepdf.PDF objects instead of files"""
     pdf_output = pikepdf.Pdf.new()
     _copy_n_transform(pdf_input, pdf_output, pages, quit_flag)
@@ -363,28 +396,93 @@ def export_doc(pdf_input, pages, mdata, files_out, quit_flag):
                 return
             outpdf = pikepdf.Pdf.new()
             _set_meta(mdata, pdf_input, outpdf)
-            try:
-                # needed to add this, probably related to pikepdf < 2.7.0 workaround
-                page = outpdf.copy_foreign(page)
-            except NotImplementedError:
-                # This is pikepdf >= 6
-                pass
             # works without make_indirect as already applied to this page
             outpdf.pages.append(page)
             _remove_unreferenced_resources(outpdf)
             outpdf.save(files_out[n])
     else:
         if isinstance(files_out[0], str):
-            _set_meta(mdata, pdf_input, pdf_output)
+            if not test_mode:
+                _set_meta(mdata, pdf_input, pdf_output)
             _remove_unreferenced_resources(pdf_output)
-        pdf_output.save(files_out[0])
+        if test_mode:
+            pdf_output.save(files_out[0], qdf=True, static_id=True, compress_streams=False,
+                            stream_decode_level=pikepdf.StreamDecodeLevel.all)
+        else:
+            pdf_output.save(files_out[0])
 
 
-def export(files, pages, mdata, files_out, quit_flag, _export_msg):
+def _add_json_entries(json: Dict[str, Any], files: List[List[str]], page: Page) -> None:
+    """Create an entry for the job json "pages" list."""
+    pages_entry = {"file": files[page.nfile - 1][0],  # copyname
+                   "range": str(page.npage)}
+    if len(files[page.nfile - 1][1]) > 0:
+        pages_entry["password"] = files[page.nfile - 1][1]
+    json["pages"].append(pages_entry)
+
+
+def _create_job(files: List[List[str]], pages: List[Page], files_out: List[str], quit_flag=None,
+                test_mode: bool = False):
+    """ Same as _copy_n_transform, except it use the pikepdf Job interface. Requires pikepdf >= 8.0 """
+    # Generate the output PDF file including temporary overlay/ underlay pages. We don't need to call
+    # _append_page as the Job interface copies pages / annotations as necessary. We can also delay getting
+    # our MediaBoxes until the transformation stage.
+    json = dict(outputFile=files_out[0], pages=[], removeUnreferencedResources="yes")
+    if test_mode:
+        json.update(qdf="", staticId="", compressStreams="n", decodeLevel="all")
+    if len(files) > 0 and len(files[0][0]) > 0:
+        json["inputFile"] = files[0][0]  # We are treating files [0] as the main document
+        if len(files[0][1]) > 0:
+            json["password"] = files[0][1]
+    else:
+        json["inputFile"] = "."
+
+    for page in pages:
+        if quit_flag is not None and quit_flag.is_set():
+            return None
+        _add_json_entries(json, files, page)
+        for lpage in page.layerpages:
+            # Layer pages are temporarily added after the page they belong to
+            _add_json_entries(json, files, lpage)
+    return pikepdf.Job(json)
+
+
+def export_doc_job(pdf_input: List[pikepdf.Pdf], files: List[List[str]], pages: List[Page], mdata, files_out: List[str],
+                   quit_flag, test_mode: bool = False) -> None:
+    """  Same as export() but uses the pikepdf Job interface. Requires pikedf >= 8.0. """
+    job = _create_job(files, pages, files_out, quit_flag, test_mode)
+    pdf_output = job.create_pdf()
+
+    _transform_job(pdf_output, pages, quit_flag)
+
+    if quit_flag is not None and quit_flag.is_set():
+        return
+    if isinstance(files_out[0], str):
+        # Only needed when saving to file, not when printing
+        mdata = metadata.merge_doc(mdata, pdf_input)
+    if len(files_out) > 1:
+        for n, page in enumerate(pdf_output.pages):
+            if quit_flag is not None and quit_flag.is_set():
+                return
+            outpdf = pikepdf.Pdf.new()
+            _set_meta(mdata, pdf_input, outpdf)
+            outpdf.pages.append(page)
+            _remove_unreferenced_resources(outpdf)
+            outpdf.save(files_out[n])
+    else:
+        if isinstance(files_out[0], str) and not test_mode:
+            _set_meta(mdata, [pdf_output], pdf_output)
+        job.write_pdf(pdf_output)
+
+
+def export(files, pages, mdata, files_out, quit_flag, _export_msg, test_mode=False):
     pdf_input = [
         pikepdf.open(copyname, password=password) for copyname, password in files
     ]
-    export_doc(pdf_input, pages, mdata, files_out, quit_flag)
+    if version.parse(pikepdf.__version__) < version.Version("8.0"):
+        export_doc(pdf_input, pages, mdata, files_out, quit_flag, test_mode)
+    else:
+        export_doc_job(pdf_input, files, pages, mdata, files_out, quit_flag, test_mode)
 
 
 def num_pages(filepath):
@@ -446,42 +544,93 @@ def generate_booklet(pdfqueue, tmp_dir, pages):
                 Contents=pikepdf.Stream(file, content_txt.encode())
             )
 
-        # workaround for pikepdf <= 2.6.0. See https://github.com/pikepdf/pikepdf/issues/174
-        if pikepdf.__version__ < '2.7.0':
-            newpage = file.make_indirect(newpage)
-        file.pages.append(newpage)
+        file.pages.append(pikepdf.Page(newpage))
     for __ in range(to_remove):
         del file.pages[0]
     file.save(filename)
     return filename
 
 
+class PrintSettingsWidget(Gtk.Grid):
+    def __init__(self, scale_mode, auto_rotate):
+        super().__init__(margin=0, row_spacing=6, column_spacing=12, border_width=12)
+        lbl = Gtk.Label(_("Scale mode:"), margin=0)
+        self.combo = Gtk.ComboBoxText(margin=0)
+        self.combo.append("NONE", _("None"))
+        self.combo.append("PRINTABLE", _("Fit to Printable Area"))
+        self.combo.append("FULL", _("Fit to Full Page"))
+        self.combo.set_active_id(scale_mode)
+        self.cb = Gtk.CheckButton(label=_("Auto Rotate"), margin=0)
+        self.cb.set_active(auto_rotate)
+        self.attach(lbl, 0, 1, 1, 1)
+        self.attach(self.combo, 1, 1, 2, 1)
+        self.attach(self.cb, 0, 2, 2, 1)
+        self.show_all()
+
+    def get_scale_mode(self):
+        return self.combo.get_active_id()
+
+    def get_auto_rotate(self):
+        return self.cb.get_active()
+
+
 # Adapted from https://stackoverflow.com/questions/28325525/python-gtk-printoperation-print-a-pdf
 class PrintOperation(Gtk.PrintOperation):
     MESSAGE=_("Printing…")
     def __init__(self, app):
-        super().__init__()
+        super().__init__(embed_page_setup=True, support_selection=True)
         self.app = app
+        self.connect("create-custom-widget", self.create_custom_widget)
+        self.connect("custom-widget-apply", self.custom_widget_apply)
+        self.connect("request-page-setup", self.request_page_setup)
         self.connect("begin-print", self.begin_print, None)
         self.connect("end-print", self.end_print, None)
         self.connect("draw-page", self.draw_page, None)
         self.connect("preview", self.preview, None)
         self.pdf_input = None
         self.message = self.MESSAGE
+        self.scale_mode = self.app.config.scale_mode()
+        self.auto_rotate = self.app.config.auto_rotate()
+        self.set_use_full_page(self.scale_mode != 'PRINTABLE')
+        self.snums = [p.get_indices() for p in reversed(app.iconview.get_selected_items())]
+        self.set_has_selection(len(self.snums) > 0)
+
+    def create_custom_widget(self, operation):
+        self.set_custom_tab_label(_("Page Handling"))
+        return PrintSettingsWidget(self.scale_mode, self.auto_rotate)
+
+    def custom_widget_apply(self, operation, widget):
+        self.scale_mode = widget.get_scale_mode()
+        self.auto_rotate = widget.get_auto_rotate()
+        self.set_use_full_page(self.scale_mode != 'PRINTABLE')
+
+    def request_page_setup(self, operation, print_ctx, page_num, setup):
+        if self.auto_rotate:
+            w_page = self.pages[page_num].width_in_points()
+            h_page = self.pages[page_num].height_in_points()
+            if w_page >= h_page:
+                setup.set_orientation(Gtk.PageOrientation.LANDSCAPE)
+            else:
+                setup.set_orientation(Gtk.PageOrientation.PORTRAIT)
 
     def preview(self, operation, preview_op, print_ctx, parent, user_data):
         self.message = _("Rendering Preview…")
 
     def begin_print(self, operation, print_ctx, print_data):
-        self.set_n_pages(len(self.app.model))
         self.app.set_export_state(True, self.message)
+        psel = self.get_print_settings().get_print_pages() == Gtk.PrintPages.SELECTION
+        nums = self.snums if psel else range(len(self.app.model))
+        self.pages = [self.app.model[n][0].duplicate(incl_thumbnail=False) for n in nums]
+        self.app.apply_hide_margins_on_pages(self.pages)
+        self.set_n_pages(len(self.pages))
+
         # Open pikepdf objects for all pages that has been modified
         nfiles = set()
-        for row in self.app.model:
-            if row[0].unmodified():
+        for p in self.pages:
+            if p.unmodified():
                 continue
-            nfiles.add(row[0].nfile)
-            for lp in row[0].layerpages:
+            nfiles.add(p.nfile)
+            for lp in p.layerpages:
                 nfiles.add(lp.nfile)
         self.pdf_input = [None] * len(self.app.pdfqueue)
         for nfile in nfiles:
@@ -491,6 +640,8 @@ class PrintOperation(Gtk.PrintOperation):
     def end_print(self, operation, print_ctx, print_data):
         self.app.set_export_state(False)
         self.message = self.MESSAGE
+        self.app.config.set_scale_mode(self.scale_mode)
+        self.app.config.set_auto_rotate(self.auto_rotate)
 
     def draw_page(self, operation, print_ctx, page_num, print_data):
         cairo_ctx = print_ctx.get_cairo_context()
@@ -498,7 +649,29 @@ class PrintOperation(Gtk.PrintOperation):
         cairo_ctx.scale(print_ctx.get_dpi_x() / 72, print_ctx.get_dpi_y() / 72)
         if page_num >= len(self.app.model):
             return
-        p = self.app.model[page_num][0]
+        setup = print_ctx.get_page_setup()
+        if self.scale_mode == 'PRINTABLE':
+            w_paper = setup.get_page_width(Gtk.Unit.POINTS)
+            h_paper = setup.get_page_height(Gtk.Unit.POINTS)
+        else:
+            w_paper = setup.get_paper_width(Gtk.Unit.POINTS)
+            h_paper = setup.get_paper_height(Gtk.Unit.POINTS)
+        w_page = self.pages[page_num].width_in_points()
+        h_page = self.pages[page_num].height_in_points()
+        print_scale = self.get_print_settings().get_scale() / 100
+        scale = 1
+        if self.scale_mode != 'NONE':
+            w_scale = w_paper / (w_page * print_scale)
+            h_scale = h_paper / (h_page * print_scale)
+            scale = min(w_scale, h_scale)
+            cairo_ctx.scale(scale, scale)
+
+        # Center page on paper
+        dx = w_paper / (scale * print_scale) - w_page
+        dy = h_paper / (scale * print_scale) - h_page
+        cairo_ctx.translate(dx / 2, dy / 2)
+
+        p = self.pages[page_num]
         if p.unmodified():
             pdfdoc = self.app.pdfqueue[p.nfile - 1]
             page = pdfdoc.document.get_page(p.npage - 1)
